@@ -84,6 +84,24 @@ Not implemented yet.
 
 This option can be used to generate normal HTML output. By default, it is ' />', which is xHTML, change to '>' for normal HTML.
 
+=item img_ids
+
+Controls if <img> tags generated have an id attribute. Defaults to true. 
+Turn off for compatibility with the original markdown.
+
+=item heading_ids
+
+Controls if <hX> tags generated have an id attribute. Defaults to true. 
+Turn off for compatibility with the original markdown.
+
+=item bibliography_title
+
+The title of the generated bibliography, defaults to 'Bibliography'.
+
+=item tab_width
+
+Controls indent width in the generated markup, defaults to 4
+
 =back
 
 A number of possible items of metadata can also be supplied as options. 
@@ -145,22 +163,20 @@ This is the base URL for referencing wiki pages. In this is not supplied, all wi
 =cut
 
 use Digest::MD5 qw(md5_hex);
+use Encode      qw();
 use Carp        qw(croak);
 use base        'Exporter';
 
-our $VERSION   = '1.0.5';
+our $VERSION   = '1.0.6';
 our @EXPORT_OK = qw(markdown);
 
 ## Disabled; causes problems under Perl 5.6.1:
 # use utf8;
 # binmode( STDOUT, ":utf8" );  # c.f.: http://acis.openlib.org/dev/perl-unicode-struggle.html
 
-### FIXME: Make these symbols configurable ###
-# $g_bibliography_title
-
 # Regex to match balanced [brackets]. See Friedl's
 # "Mastering Regular Expressions", 2nd Ed., pp. 328-331.
-my $g_nested_brackets; 
+my ($g_nested_brackets, $g_nested_parens); 
 $g_nested_brackets = qr{
     (?>                                 # Atomic matching
        [^\[\]]+                         # Anything other than brackets
@@ -170,7 +186,16 @@ $g_nested_brackets = qr{
        \]
     )*
 }x;
-
+# Doesn't allow for whitespace, because we're using it to match URLs:
+$g_nested_parens = qr{
+	(?> 								# Atomic matching
+	   [^()\s]+							# Anything other than parens or whitespace
+	 | 
+	   \(
+		 (??{ $g_nested_parens })		# Recursive set of nested brackets
+	   \)
+	)*
+}x;
 
 # Table of hash values for escaped characters:
 my %g_escape_table;
@@ -180,19 +205,7 @@ foreach my $char (split //, '\\`*_{}[]()>#+-.!') {
 
 # Global hashes, used by various utility routines
 # FIXME - to be moved into instance data!
-my %g_urls = ();
-my %g_titles= ();
-my %g_html_blocks = ();
 my %g_metadata_newline = ();
-my %g_crossrefs = ();
-my %g_footnotes = ();
-my %g_attributes = ();
-my @g_used_footnotes = ();
-
-my $g_citation_counter = 0;
-my @g_used_references = ();
-my %g_references = ();
-my $g_bibliography_title = "Bibliography";
 
 # FIXME - If we're using metadata, newlines (in the metadata have to be \n). Make configurable.
 $g_metadata_newline{default} = "\n";
@@ -223,13 +236,16 @@ sub new {
     
     $p{empty_element_suffix} ||= ' />'; # Change to ">" for HTML output
     
+    #$p{heading_ids} = defined $p{heading_ids} ? $p{heading_ids} : 1;
+    
     # For use with WikiWords and [[Wiki Links]]
     # NOTE: You can use \WikiWord to prevent a WikiWord from being treated as a link
     $p{use_wikilinks} = $p{use_wikilinks} ? 1 : 0;
     
-    # Used to track when we're inside an ordered or unordered list
-    # (see _ProcessListItems() for details):
-    $p{_list_level} = 0;
+    $p{heading_ids} = defined $p{heading_ids} ? $p{heading_ids} : 1;
+    $p{img_ids}     = defined $p{img_ids}     ? $p{img_ids}     : 1;
+    
+    $p{bibliography_title} ||= 'Bibliography'; # FIXME - Test and document, can also be in metadata!
     
     my $self = { %p };
     bless $self, ref($class) || $class;
@@ -278,17 +294,19 @@ sub markdown {
     # from other articles when generating a page which contains more than
     # one article (e.g. an index page that shows the N most recent
     # articles):
-    %g_urls            = ();
-    %g_titles          = ();
-    %g_html_blocks     = ();
-    %g_crossrefs       = ();
-    %g_footnotes       = ();
-    @g_used_footnotes  = ();
-    @g_used_references = ();
-
-    %g_urls        = %{$options->{urls}}        if $options->{urls};
-    %g_titles      = %{$options->{titles}}      if $options->{titles};
-    %g_html_blocks = %{$options->{html_blocks}} if $options->{html_blocks};
+    $self->{_urls}        = $options->{urls}        ? $options->{urls}        : {}; # FIXME - document and test passing this option.
+    $self->{_titles}      = $options->{titles}      ? $options->{titles}      : {}; # FIXME - ditto
+    $self->{_html_blocks} = $options->{html_blocks} ? $options->{html_blocks} : {}; # FIXME - ditto
+    $self->{_crossrefs}   = {};
+    $self->{_footnotes}   = {};
+    $self->{_references}  = {};
+    $self->{_attributes}  = {}; # What is this used for again?
+    $self->{_used_footnotes}  = []; # Why do we need 2 data structures for footnotes? FIXME
+    $self->{_used_references} = []; # Ditto for references
+    $self->{_citation_counter} = 0;
+    # Used to track when we're inside an ordered or unordered list
+    # (see _ProcessListItems() for details)
+    $self->{_list_level} = 0;
 
     my $t =  $self->_Markdown($text);
 
@@ -325,9 +343,6 @@ sub _Markdown {
     # match consecutive blank lines with /\n+/ instead of something
     # contorted like /[ \t]*\n+/ .
     $text =~ s/^[ \t]+$//mg;
-
-    # Strip leading blank lines
-    $text =~ s/^\n+//s;
     
     # Strip out MetaData
     $text = $self->_ParseMetaData($text) if $self->{use_metadata};
@@ -411,15 +426,15 @@ sub _StripLinkDefinitions {
                         (?:\n+|\Z)
                     }
                     {}mx) {
-        $g_urls{lc $1} = $self->_EncodeAmpsAndAngles( $2 );    # Link IDs are case-insensitive
+        $self->{_urls}{lc $1} = $self->_EncodeAmpsAndAngles( $2 );    # Link IDs are case-insensitive
         if ($3) {
-            $g_titles{lc $1} = $3;
-            $g_titles{lc $1} =~ s/"/&quot;/g;
+            $self->{_titles}{lc $1} = $3;
+            $self->{_titles}{lc $1} =~ s/"/&quot;/g;
         }
         
         # MultiMarkdown addition "
         if ($4) {
-            $g_attributes{lc $1} = $4;
+            $self->{_attributes}{lc $1} = $4;
         }
         # /addition
     }
@@ -427,114 +442,161 @@ sub _StripLinkDefinitions {
     return $text;
 }
 
+sub _md5_utf8 {
+   # Internal function used to safely MD5sum chunks of the input, which might be Unicode in Perl's internal representation.
+   my $input = shift;
+   return unless defined $input;
+   if (Encode::is_utf8 $input) {
+       return md5_hex(Encode::encode('utf8', $input));
+    } 
+    else {
+        return md5_hex($input);
+    }
+}
 
 sub _HashHTMLBlocks {
     my ($self, $text) = @_;
     my $less_than_tab = $self->{tab_width} - 1;
 
-    # Hashify HTML blocks:
-    # We only want to do this for block-level HTML tags, such as headers,
-    # lists, and tables. That's because we still want to wrap <p>s around
-    # "paragraphs" that are wrapped in non-block-level tags, such as anchors,
-    # phrase emphasis, and spans. The list of tags we're looking for is
-    # hard-coded:
-    my $block_tags_a = qr/p|div|h[1-6]|blockquote|pre|table|dl|ol|ul|script|noscript|form|fieldset|iframe|math|ins|del/;
-    my $block_tags_b = qr/p|div|h[1-6]|blockquote|pre|table|dl|ol|ul|script|noscript|form|fieldset|iframe|math/;
+	# Hashify HTML blocks:
+	# We only want to do this for block-level HTML tags, such as headers,
+	# lists, and tables. That's because we still want to wrap <p>s around
+	# "paragraphs" that are wrapped in non-block-level tags, such as anchors,
+	# phrase emphasis, and spans. The list of tags we're looking for is
+	# hard-coded:
+	my $block_tags = qr{
+		  (?:
+			p         |  div     |  h[1-6]  |  blockquote  |  pre       |  table  |
+			dl        |  ol      |  ul      |  script      |  noscript  |  form   |
+			fieldset  |  iframe  |  math    |  ins         |  del
+		  )
+		}x;
 
-    # First, look for nested blocks, e.g.:
-    #   <div>
-    #       <div>
-    #       tags for inner block must be indented.
-    #       </div>
-    #   </div>
-    #
-    # The outermost tags must start at the left margin for this to match, and
-    # the inner nested divs must be indented.
-    # We need to do this before the next, more liberal match, because the next
-    # match will start at the first `<div>` and stop at the first `</div>`.
-    $text =~ s{
-                (                       # save in $1
-                    ^                   # start of line  (with /m)
-                    <($block_tags_a)    # start tag = $2
-                    \b                  # word break
-                    (.*\n)*?            # any number of lines, minimally matching
-                    </\2>               # the matching end tag
-                    [ \t]*              # trailing spaces/tabs
-                    (?=\n+|\Z)  # followed by a newline or end of document
-                )
-            }{
-                my $key = md5_hex($1);
-                $g_html_blocks{$key} = $1;
-                "\n\n" . $key . "\n\n";
-            }egmx;
+	my $tag_attrs = qr{
+						(?:				# Match one attr name/value pair
+							\s+				# There needs to be at least some whitespace
+											# before each attribute name.
+							[\w.:_-]+		# Attribute name
+							\s*=\s*
+							(?:
+								".+?"		# "Attribute value"
+							 |
+								'.+?'		# 'Attribute value'
+							)
+						)*				# Zero or more
+					}x;
+
+	my $empty_tag = qr{< \w+ $tag_attrs \s* />}xms;
+	my $open_tag =  qr{< $block_tags $tag_attrs \s* >}xms;
+	my $close_tag = undef;	# let Text::Balanced handle this
+
+	use Text::Balanced qw(gen_extract_tagged);
+	my $extract_block = gen_extract_tagged($open_tag, $close_tag, undef, { ignore => [$empty_tag] });
+
+	my @chunks;
+	## TO-DO: the 0,3 on the next line ought to respect the
+	## tabwidth, or else, we should mandate 4-space tabwidth and
+	## be done with it:
+	while ($text =~ s{^(([ ]{0,3}<)?.*\n)}{}m) {
+		my $cur_line = $1;
+		if (defined $2) {
+			# current line could be start of code block
+
+			my ($tag, $remainder) = $extract_block->($cur_line . $text);
+			if ($tag) {
+				my $key = _md5_utf8($tag);
+				$self->{_html_blocks}{$key} = $tag;
+				push @chunks, "\n\n" . $key . "\n\n";
+				$text = $remainder;
+			}
+			else {
+				# No tag match, so toss $cur_line into @chunks
+				push @chunks, $cur_line;
+			}
+		}
+		else {
+			# current line could NOT be start of code block
+			push @chunks, $cur_line;
+		}
+
+	}
+	push @chunks, $text; # Whatever is left.
+
+	$text = join '', @chunks;
 
 
-    #
-    # Now match more liberally, simply from `\n<tag>` to `</tag>\n`
-    #
-    $text =~ s{
-                (                       # save in $1
-                    ^                   # start of line  (with /m)
-                    <($block_tags_b)    # start tag = $2
-                    \b                  # word break
-                    (.*\n)*?            # any number of lines, minimally matching
-                    .*</\2>             # the matching end tag
-                    [ \t]*              # trailing spaces/tabs
-                    (?=\n+|\Z)  # followed by a newline or end of document
-                )
-            }{
-                my $key = md5_hex($1);
-                $g_html_blocks{$key} = $1;
-                "\n\n" . $key . "\n\n";
-            }egmx;
-    # Special case just for <hr />. It was easier to make a special case than
-    # to make the other regex more complicated. 
-    $text =~ s{
-                (?:
-                    (?<=\n\n)       # Starting after a blank line
-                    |               # or
-                    \A\n?           # the beginning of the doc
-                )
-                (                       # save in $1
-                    [ ]{0,$less_than_tab}
-                    <(hr)               # start tag = $2
-                    \b                  # word break
-                    ([^<>])*?           # 
-                    /?>                 # the matching end tag
-                    [ \t]*
-                    (?=\n{2,}|\Z)       # followed by a blank line or end of document
-                )
-            }{
-                my $key = md5_hex($1);
-                $g_html_blocks{$key} = $1;
-                "\n\n" . $key . "\n\n";
-            }egx;
 
-    # Special case for standalone HTML comments:
-    $text =~ s{
-                (?:
-                    (?<=\n\n)       # Starting after a blank line
-                    |               # or
-                    \A\n?           # the beginning of the doc
-                )
-                (                       # save in $1
-                    [ ]{0,$less_than_tab}
-                    (?s:
-                        <!
-                        (--.*?--\s*)+
-                        >
-                    )
-                    [ \t]*
-                    (?=\n{2,}|\Z)       # followed by a blank line or end of document
-                )
-            }{
-                my $key = md5_hex($1);
-                $g_html_blocks{$key} = $1;
-                "\n\n" . $key . "\n\n";
-            }egx;
+	# Special case just for <hr />. It was easier to make a special case than
+	# to make the other regex more complicated.	
+	$text =~ s{
+				(?:
+					(?<=\n\n)		# Starting after a blank line
+					|				# or
+					\A\n?			# the beginning of the doc
+				)
+				(						# save in $1
+					[ ]{0,$less_than_tab}
+					<(hr)				# start tag = $2
+					\b					# word break
+					([^<>])*?			# 
+					/?>					# the matching end tag
+					[ \t]*
+					(?=\n{2,}|\Z)		# followed by a blank line or end of document
+				)
+			}{
+				my $key = _md5_utf8($1);
+				$self->{_html_blocks}{$key} = $1;
+				"\n\n" . $key . "\n\n";
+			}egx;
+
+	# Special case for standalone HTML comments:
+	$text =~ s{
+				(?:
+					(?<=\n\n)		# Starting after a blank line
+					|				# or
+					\A\n?			# the beginning of the doc
+				)
+				(						# save in $1
+					[ ]{0,$less_than_tab}
+					(?s:
+						<!
+						(--.*?--\s*)+
+						>
+					)
+					[ \t]*
+					(?=\n{2,}|\Z)		# followed by a blank line or end of document
+				)
+			}{
+				my $key = _md5_utf8($1);
+				$self->{_html_blocks}{$key} = $1;
+				"\n\n" . $key . "\n\n";
+			}egx;
+
+	# PHP and ASP-style processor instructions (<?…?> and <%…%>)
+	$text =~ s{
+				(?:
+					(?<=\n\n)		# Starting after a blank line
+					|				# or
+					\A\n?			# the beginning of the doc
+				)
+				(						# save in $1
+					[ ]{0,$less_than_tab}
+					(?s:
+						<([?%])			# $2
+						.*?
+						\2>
+					)
+					[ \t]*
+					(?=\n{2,}|\Z)		# followed by a blank line or end of document
+				)
+			}{
+				my $key = _md5_utf8($1);
+				$self->{_html_blocks}{$key} = $1;
+				"\n\n" . $key . "\n\n";
+			}egx;
 
 
-    return $text;
+	return $text;
 }
 
 
@@ -593,7 +655,7 @@ sub _RunSpanGamut {
     my ($self, $text) = @_;
 
     $text = $self->_DoCodeSpans($text);
-
+	$text = $self->_EscapeSpecialCharsWithinTagAttributes($text);
     $text = $self->_EscapeSpecialChars($text);
 
     # Process anchor and image tags. Images must come first,
@@ -654,6 +716,30 @@ sub _EscapeSpecialChars {
     return $text;
 }
 
+sub _EscapeSpecialCharsWithinTagAttributes {
+#
+# Within tags -- meaning between < and > -- encode [\ ` * _] so they
+# don't conflict with their use in Markdown for code, italics and strong.
+# We're replacing each such character with its corresponding MD5 checksum
+# value; this is likely overkill, but it should prevent us from colliding
+# with the escape values by accident.
+#
+	my ($self, $text) = @_;
+	my $tokens ||= $self->_TokenizeHTML($text);
+	$text = '';   # rebuild $text from the tokens
+
+	foreach my $cur_token (@$tokens) {
+		if ($cur_token->[0] eq "tag") {
+			$cur_token->[1] =~  s! \\ !$g_escape_table{'\\'}!gx;
+			$cur_token->[1] =~  s{ (?<=.)</?code>(?=.)  }{$g_escape_table{'`'}}gx;
+			$cur_token->[1] =~  s! \* !$g_escape_table{'*'}!gx;
+			$cur_token->[1] =~  s! _  !$g_escape_table{'_'}!gx;
+		}
+		$text .= $cur_token->[1];
+	}
+	return $text;
+}
+
 
 sub _DoAnchors {
 #
@@ -681,6 +767,7 @@ sub _DoAnchors {
         my $result;
         my $whole_match = $1;
         my $link_text   = $2;
+        #warn("Got match $1 text $2");
         my $link_id     = lc $3;
 
         if ($link_id eq "") {
@@ -689,13 +776,13 @@ sub _DoAnchors {
 
         # Allow automatic cross-references to headers
         my $label = $self->Header2Label($link_id);
-        if (defined $g_crossrefs{$label}) {
-            my $url = $g_crossrefs{$label};
+        if (defined $self->{_crossrefs}{$label}) {
+            my $url = $self->{_crossrefs}{$label};
             $url =~ s! \* !$g_escape_table{'*'}!gx;     # We've got to encode these to avoid
             $url =~ s!  _ !$g_escape_table{'_'}!gx;     # conflicting with italics/bold.
             $result = qq[<a href="$url"];
-            if ( defined $g_titles{$label} ) {
-                my $title = $g_titles{$label};
+            if ( defined $self->{_titles}{$label} ) {
+                my $title = $self->{_titles}{$label};
                 $title =~ s! \* !$g_escape_table{'*'}!gx;
                 $title =~ s!  _ !$g_escape_table{'_'}!gx;
                 $result .=  " title=\"$title\"";
@@ -703,16 +790,16 @@ sub _DoAnchors {
             $result .= $self->_DoAttributes($label);
             $result .= ">$link_text</a>";
         } 
-        elsif (defined $g_urls{$link_id}) {
-            my $url = $g_urls{$link_id};
+        elsif (defined $self->{_urls}{$link_id}) {
+            my $url = $self->{_urls}{$link_id};
             $url =~ s! \* !$g_escape_table{'*'}!gx;     # We've got to encode these to avoid
             $url =~ s!  _ !$g_escape_table{'_'}!gx;     # conflicting with italics/bold.
             $result = "<a href=\"$url\"";
-            if ( defined $g_titles{$link_id} ) {
-                my $title = $g_titles{$link_id};
+            if ( defined $self->{_titles}{$link_id} ) {
+                my $title = $self->{_titles}{$link_id};
                 $title =~ s! \* !$g_escape_table{'*'}!gx;
                 $title =~ s!  _ !$g_escape_table{'_'}!gx;
-                $result .=  " title=\"$title\"";
+                $result .=  qq{ title="$title"};
             }
             $result .= $self->_DoAttributes($label);
             $result .= ">$link_text</a>";
@@ -733,12 +820,13 @@ sub _DoAnchors {
           \]
           \(            # literal paren
             [ \t]*
-            <?(.*?)>?   # href = $3
+            ($g_nested_parens)   # href = $3
             [ \t]*
             (           # $4
               (['"])    # quote char = $5
               (.*?)     # Title = $6
               \5        # matching quote
+              [ \t]*	# ignore any spaces/tabs between closing quote and )
             )?          # title is optional
           \)
         )
@@ -751,6 +839,7 @@ sub _DoAnchors {
         
         $url =~ s! \* !$g_escape_table{'*'}!gx;     # We've got to encode these to avoid
         $url =~ s!  _ !$g_escape_table{'_'}!gx;     # conflicting with italics/bold.
+        $url =~ s{^<(.*)>$}{$1};					# Remove <>'s surrounding URL, if present
         $result = "<a href=\"$url\"";
 
         if (defined $title) {
@@ -764,6 +853,42 @@ sub _DoAnchors {
 
         $result;
     }xsge;
+    
+    #
+	# Last, handle reference-style shortcuts: [link text]
+	# These must come last in case you've also got [link test][1]
+	# or [link test](/foo)
+	#
+	$text =~ s{
+		(					# wrap whole match in $1
+		  \[
+		    ([^\[\]]+)		# link text = $2; can't contain '[' or ']'
+		  \]
+		)
+	}{
+		my $result;
+		my $whole_match = $1;
+		my $link_text   = $2;
+		(my $link_id = lc $2) =~ s{[ ]?\n}{ }g; # lower-case and turn embedded newlines into spaces
+
+		if (defined $self->{_urls}{$link_id}) {
+			my $url = $self->{_urls}{$link_id};
+			$url =~ s! \* !$g_escape_table{'*'}!gx;		# We've got to encode these to avoid
+			$url =~ s!  _ !$g_escape_table{'_'}!gx;		# conflicting with italics/bold.
+			$result = "<a href=\"$url\"";
+			if ( defined $self->{titles}{$link_id} ) {
+				my $title = $self->{titles}{$link_id};
+				$title =~ s! \* !$g_escape_table{'*'}!gx;
+				$title =~ s!  _ !$g_escape_table{'_'}!gx;
+				$result .=  qq{ title="$title"};
+			}
+			$result .= ">$link_text</a>";
+		}
+		else {
+			$result = $whole_match;
+		}
+		$result;
+	}xsge;
 
     return $text;
 }
@@ -803,23 +928,24 @@ sub _DoImages {
         }
 
         $alt_text =~ s/"/&quot;/g;
-        if (defined $g_urls{$link_id}) {
-            my $url = $g_urls{$link_id};
+        if (defined $self->{_urls}{$link_id}) {
+            my $url = $self->{_urls}{$link_id};
             $url =~ s! \* !$g_escape_table{'*'}!gx;     # We've got to encode these to avoid
             $url =~ s!  _ !$g_escape_table{'_'}!gx;     # conflicting with italics/bold.
             
             my $label = $self->Header2Label($alt_text);
-            $g_crossrefs{$label} = "#$label";
-            if (! defined $g_titles{$link_id}) {
-                $g_titles{$link_id} = $alt_text;
-            }
+            $self->{_crossrefs}{$label} = "#$label";
+            #if (! defined $self->{_titles}{$link_id}) {
+            #    $self->{_titles}{$link_id} = $alt_text;
+            #}
             
-            $result = qq{<img id="$label" src="$url" alt="$alt_text"};
-            if (defined $g_titles{$link_id}) {
-                my $title = $g_titles{$link_id};
+            $label = $self->{img_ids} ? qq{ id="$label"} : '';
+            $result = qq{<img$label src="$url" alt="$alt_text"};
+            if (length $link_id && defined $self->{_titles}{$link_id}) {
+                my $title = $self->{_titles}{$link_id};
                 $title =~ s! \* !$g_escape_table{'*'}!gx;
                 $title =~ s!  _ !$g_escape_table{'_'}!gx;
-                $result .=  " title=\"$title\"";
+                $result .=  qq{ title="$title"};
             }
             $result .= $self->_DoAttributes($link_id);
             $result .= $self->{empty_element_suffix};
@@ -844,7 +970,7 @@ sub _DoImages {
           \]
           \(            # literal paren
             [ \t]*
-            <?(\S+?)>?  # src url = $3
+            ($g_nested_parens)  # src url - href = $3
             [ \t]*
             (           # $4
               (['"])    # quote char = $5
@@ -868,19 +994,21 @@ sub _DoImages {
         $title    =~ s/"/&quot;/g;
         $url =~ s! \* !$g_escape_table{'*'}!gx;     # We've got to encode these to avoid
         $url =~ s!  _ !$g_escape_table{'_'}!gx;     # conflicting with italics/bold.
+        $url =~ s{^<(.*)>$}{$1};					# Remove <>'s surrounding URL, if present
 
         my $label = $self->Header2Label($alt_text);
-        $g_crossrefs{$label} = "#$label";
-#       $g_titles{$label} = $alt_text;          # FIXME - I think this line should not be here
-            
-        $result = qq{<img id="$label" src="$url" alt="$alt_text"};
+        $self->{_crossrefs}{$label} = "#$label";
+#       $self->{_titles}{$label} = $alt_text;          # FIXME - I think this line should not be here
+
+        $label = $self->{img_ids} ? qq{ id="$label"} : '';
+        $result = qq{<img$label src="$url" alt="$alt_text"};
         if (defined $title) {
             $title =~ s! \* !$g_escape_table{'*'}!gx;
             $title =~ s!  _ !$g_escape_table{'_'}!gx;
             $result .=  qq{ title="$title"};
         }
         $result .= $self->{empty_element_suffix};
-
+        
         $result;
     }xsge;
 
@@ -904,23 +1032,27 @@ sub _DoHeaders {
     #     --------
     #
     $text =~ s{ ^(.+)[ \t]*\n=+[ \t]*\n+ }{
-        $label = $self->Header2Label($1);
+        $label = $self->{heading_ids} ? q{ id="} . $self->Header2Label($1) . q{"} : '';
         $header = $self->_RunSpanGamut($1);
         
-        $g_crossrefs{$label} = "#$label";
-        $g_titles{$label} = $header;
+        if ($label ne '') {
+            $self->{_crossrefs}{$label} = "#$label";
+            $self->{_titles}{$label} = $header;
+        }
         
-        "<h1 id=\"$label\">"  .  $self->_RunSpanGamut($1)  .  "</h1>\n\n";
+        "<h1$label>"  .  $self->_RunSpanGamut($1)  .  "</h1>\n\n";
     }egmx;
 
     $text =~ s{ ^(.+)[ \t]*\n-+[ \t]*\n+ }{
-        $label = $self->Header2Label($1);
+        $label = $self->{heading_ids} ? q{ id="} . $self->Header2Label($1) . q{"} : '';
         $header = $self->_RunSpanGamut($1);
         
-        $g_crossrefs{$label} = "#$label";
-        $g_titles{$label} = $header;
+        if ($label ne '') {
+            $self->{_crossrefs}{$label} = "#$label";
+            $self->{_titles}{$label} = $header;
+        }
         
-        "<h2 id=\"$label\">"  .  $self->_RunSpanGamut($1)  .  "</h2>\n\n";
+        "<h2$label>"  .  $self->_RunSpanGamut($1)  .  "</h2>\n\n";
     }egmx;
 
 
@@ -931,6 +1063,7 @@ sub _DoHeaders {
     #   ...
     #   ###### Header 6
     #
+    my $l;
     $text =~ s{
             ^(\#{1,6})  # $1 = string of #'s
             [ \t]*
@@ -940,12 +1073,15 @@ sub _DoHeaders {
             \n+
         }{
             my $h_level = length($1);
-            $label = $self->Header2Label($2);
             $header = $self->_RunSpanGamut($2);
             
-            $g_crossrefs{$label} = "#$label";
-            $g_titles{$label} = $header;
-            "<h$h_level id=\"$label\">"  .  $header  .  "</h$h_level>\n\n";
+            if ($self->{heading_ids}) {
+                $label = $self->Header2Label($2);
+                $self->{_crossrefs}{$label} = "#$label";
+                $self->{_titles}{$label} = $header;
+            }
+            $l = $self->{heading_ids} ? $h_level . ' id="' . $label . '"' : $h_level;
+            "<h$l>"  .  $header  .  "</h$h_level>\n\n";
         }egmx;
 
     return $text;
@@ -1114,30 +1250,30 @@ sub _DoCodeBlocks {
 
     my ($self, $text) = @_;
 
-    $text =~ s{
-            (?:\n\n|\A)
-            (               # $1 = the code block -- one or more lines, starting with a space/tab
-              (?:
-                (?:[ ]{$self->{tab_width}} | \t)  # Lines must start with a tab or a tab-width of spaces
-                .*\n+
-              )+
-            )
-            ((?=^[ ]{0,$self->{tab_width}}\S)|\Z) # Lookahead for non-space at line-start, or end of doc
-        }{
-            my $codeblock = $1;
-            my $result; # return value
+ 	$text =~ s{
+		(?:\n\n|\A)
+		(	            # $1 = the code block -- one or more lines, starting with a space/tab
+		  (?:
+		    (?:[ ]{$self->{tab_width}} | \t)  # Lines must start with a tab or a tab-width of spaces
+		    .*\n+
+		  )+
+		)
+		((?=^[ ]{0,$self->{tab_width}}\S)|\Z)	# Lookahead for non-space at line-start, or end of doc
+	}{
+    	my $codeblock = $1;
+    	my $result; # return value
 
-            $codeblock = $self->_EncodeCode($self->_Outdent($codeblock));
-            $codeblock = $self->_Detab($codeblock);
-            $codeblock =~ s/\A\n+//; # trim leading newlines
-            $codeblock =~ s/\s+\z//; # trim trailing whitespace
+    	$codeblock = $self->_EncodeCode($self->_Outdent($codeblock));
+    	$codeblock = $self->_Detab($codeblock);
+    	$codeblock =~ s/\A\n+//; # trim leading newlines
+    	$codeblock =~ s/\n+\z//; # trim trailing newlines
 
-            $result = "\n\n<pre><code>" . $codeblock . "\n</code></pre>\n\n";
+    	$result = "\n\n<pre><code>" . $codeblock . "\n</code></pre>\n\n";
 
-            $result;
-        }egmx;
+    	$result;
+	}egmx;
 
-    return $text;
+	return $text;
 }
 
 
@@ -1169,19 +1305,20 @@ sub _DoCodeSpans {
 
     my ($self, $text) = @_;
 
-    $text =~ s@
-            (`+)        # $1 = Opening run of `
-            (.+?)       # $2 = The code block
-            (?<!`)
-            \1          # Matching closer
-            (?!`)
-        @
-            my $c = "$2";
-            $c =~ s/^[ \t]*//g; # leading whitespace
-            $c =~ s/[ \t]*$//g; # trailing whitespace
-            $c = $self->_EncodeCode($c);
-            "<code>$c</code>";
-        @egsx;
+	$text =~ s@
+			(?<!\\)		# Character before opening ` can't be a backslash
+			(`+)		# $1 = Opening run of `
+			(.+?)		# $2 = The code block
+			(?<!`)
+			\1			# Matching closer
+			(?!`)
+		@
+ 			my $c = "$2";
+ 			$c =~ s/^[ \t]*//g; # leading whitespace
+ 			$c =~ s/[ \t]*$//g; # trailing whitespace
+ 			$c = $self->_EncodeCode($c);
+			"<code>$c</code>";
+		@egsx;
 
     return $text;
 }
@@ -1314,7 +1451,7 @@ sub _FormParagraphs {
     # Wrap <p> tags.
     #
     foreach (@grafs) {
-        unless (defined( $g_html_blocks{$_} )) {
+        unless (defined( $self->{_html_blocks}{$_} )) {
             $_ = $self->_RunSpanGamut($_);
             s/^([ \t]*)/<p>/;
             $_ .= "</p>";
@@ -1325,8 +1462,8 @@ sub _FormParagraphs {
     # Unhashify HTML blocks
     #
     foreach (@grafs) {
-        if (defined( $g_html_blocks{$_} )) {
-            $_ = $g_html_blocks{$_};
+        if (defined( $self->{_html_blocks}{$_} )) {
+            $_ = $self->{_html_blocks}{$_};
         }
     }
 
@@ -1426,7 +1563,6 @@ sub _EncodeEmailAddress {
 
     my ($self, $addr) = @_;
 
-    srand; # FIXME - NOT SMART TO DO THIS MULTIPLE TIMES!!!
     my @encode = (
         sub { '&#' .                 ord(shift)   . ';' },
         sub { '&#x' . sprintf( "%X", ord(shift) ) . ';' },
@@ -1548,7 +1684,7 @@ sub _ParseMetaData { # FIXME - This is really really ugly!
     my ($inMetaData, $currentKey) = (1, '');
     
     foreach my $line ( split /\n/, $text ) {
-        $line =~ /^$/ and $inMetaData = 0 and $clean_text .= $line and next;
+        $line =~ /^\s*$/ and $inMetaData = 0 and $clean_text .= $line and next;
         if ($inMetaData) {
             if ($line =~ /^([a-zA-Z0-9][0-9a-zA-Z _-]+?):\s*(.*)$/ ) {
                 $currentKey = $1;
@@ -1604,7 +1740,7 @@ sub _StripFootnoteDefinitions {
         my $footnote = "$2\n";
         $footnote =~ s/^[ ]{0,$self->{tab_width}}//gm;
     
-        $g_footnotes{$self->Header2Label($id)} = $footnote;
+        $self->{_footnotes}{$self->Header2Label($id)} = $footnote;
     }
     
     return $text;
@@ -1614,14 +1750,14 @@ sub _DoFootnotes {
     my ($self, $text) = @_;
     
     # First, run routines that get skipped in footnotes
-    foreach my $label (sort keys %g_footnotes) {
-        my $footnote = $self->_RunBlockGamut($g_footnotes{$label});
+    foreach my $label (sort keys %{ $self->{_footnotes} }) {
+        my $footnote = $self->_RunBlockGamut($self->{_footnotes}{$label});
 
         # strip leading and trailing <p> tags (they will be added later)
         $footnote =~ s/^<p>//s;
         $footnote =~ s/<\/p>$//s;
         $footnote = $self->_DoMarkdownCitations($footnote);
-        $g_footnotes{$label} = $footnote;
+        $self->{_footnotes}{$label} = $footnote;
     }
     
     my $footnote_counter = 0;
@@ -1632,11 +1768,11 @@ sub _DoFootnotes {
         my $result;
         my $id = $self->Header2Label($1);
         
-        if (defined $g_footnotes{$id} ) {
-#           $result = "<footnote>$g_footnotes{$self->Header2Label($id)}</footnote>"
+        if (defined $self->{_footnotes}{$id} ) {
+#           $result = "<footnote>$self->{_footnotes}{$self->Header2Label($id)}</footnote>"
             $footnote_counter++;
             $result = qq{<a href="#$id" id="f$id" class="footnote">$footnote_counter</a>};
-            push (@g_used_footnotes,$id);
+            push (@{ $self->{_used_footnotes} }, $id);
         }
         $result;
     }xsge;
@@ -1657,10 +1793,10 @@ sub _PrintFootnotes {
     my $footnote_counter = 0;
     my $result;
     
-    foreach my $id (@g_used_footnotes) {
+    foreach my $id (@{ $self->{_used_footnotes} }) {
         $footnote_counter++;
-        #$result .= qq[<div id="$id"><p><a href="#f$id" class="reversefootnote">$footnote_counter.</a> $g_footnotes{$id}</div>\n\n];
-        $result .= qq[<li id="$id"><p>$g_footnotes{$id}<a href="#f$id" class="reversefootnote">&#160;&#8617;</a></p></li>\n\n];
+        #$result .= qq[<div id="$id"><p><a href="#f$id" class="reversefootnote">$footnote_counter.</a> $self->{_footnotes}{$id}</div>\n\n];
+        $result .= qq[<li id="$id"><p>$self->{_footnotes}{$id}<a href="#f$id" class="reversefootnote">&#160;&#8617;</a></p></li>\n\n];
     }
 
     if ($footnote_counter > 0) {
@@ -1887,8 +2023,8 @@ sub _DoTables {
                 my $table_id = $self->Header2Label($3);
                 $result .= qq{<caption id="$table_id">} . $self->_RunSpanGamut($1). "</caption>\n";
                 
-                $g_crossrefs{$table_id} = "#$table_id";
-                $g_titles{$table_id} = "$1";
+                $self->{_crossrefs}{$table_id} = "#$table_id";
+                $self->{_titles}{$table_id} = "$1";
             } 
             else {
                 $result .= "<caption>" . $self->_RunSpanGamut($1). "</caption>\n";
@@ -2043,11 +2179,11 @@ sub _DoAttributes {
     my ($self, $id) = @_;
     my $result = "";
     
-    if (defined $g_attributes{$id}) {
-        while ($g_attributes{$id} =~ s/(\S+)="(.*?)"//) {
+    if (defined $self->{_attributes}{$id}) {
+        while ($self->{_attributes}{$id} =~ s/(\S+)="(.*?)"//) {
             $result .= " $1=\"$2\"";
         }
-        while ($g_attributes{$id} =~ /(\S+)=(\S+)/g) {
+        while ($self->{_attributes}{$id} =~ /(\S+)=(\S+)/g) {
             $result .= " $1=\"$2\"";
         }
     }
@@ -2079,7 +2215,7 @@ sub _StripMarkdownReferences {
         $reference =~ s/^\<p\>//s;
         $reference =~ s/\<\/p\>\s*$//s;
         
-        $g_references{$id} = $reference;
+        $self->{_references}{$id} = $reference;
     }
     
     return $text;
@@ -2099,19 +2235,18 @@ sub _DoMarkdownCitations {
         my $id = $2;
         my $count;
         
-        if (defined $g_references{$id} ) {
+        if (defined $self->{_references}{$id} ) {
             my $citation_counter=0;
             
             # See if citation has been used before
-            foreach my $old_id (@g_used_references) {
+            foreach my $old_id (@{ $self->{_used_references} }) {
                 $citation_counter++;
                 $count = $citation_counter if ($old_id eq $id);
             }
     
             if (! defined $count) {
-                $g_citation_counter++;
-                $count = $g_citation_counter;
-                push (@g_used_references,$id);
+                $count = ++$self->{_citation_counter};
+                push (@{ $self->{_used_references} }, $id);
             }
             
             $result = qq[<span class="markdowncitation"> (<a href="#$id">$count</a>];
@@ -2148,14 +2283,14 @@ sub _PrintMarkdownBibliography {
     my $citation_counter = 0;
     my $result;
     
-    foreach my $id (@g_used_references) {
+    foreach my $id (@{ $self->{_used_references} }) {
         $citation_counter++;
-        $result .= qq|<div id="$id"><p>[$citation_counter] <span class="item">$g_references{$id}</span></p></div>\n\n|;
+        $result .= qq|<div id="$id"><p>[$citation_counter] <span class="item">$self->{_references}{$id}</span></p></div>\n\n|;
     }
     $result .= "</div>";
 
     if ($citation_counter > 0) {
-        $result = qq[\n\n<div class="bibliography">\n<hr$self->{empty_element_suffix}\n<p>$g_bibliography_title</p>\n\n] . $result;
+        $result = qq[\n\n<div class="bibliography">\n<hr$self->{empty_element_suffix}\n<p>$self->{bibliography_title}</p>\n\n] . $result;
     } 
     else {
         $result = "";
@@ -2195,9 +2330,9 @@ sub _GenerateImageCrossRefs {
         }
 
         $alt_text =~ s/"/&quot;/g;
-        if (defined $g_urls{$link_id}) {
+        if (defined $self->{_urls}{$link_id}) {
             my $label = $self->Header2Label($alt_text);
-            $g_crossrefs{$label} = "#$label";
+            $self->{_crossrefs}{$label} = "#$label";
         }
         else {
             # If there's no such link ID, leave intact:
@@ -2235,7 +2370,7 @@ sub _GenerateImageCrossRefs {
 
         $alt_text =~ s/"/&quot;/g;
         my $label = $self->Header2Label($alt_text);
-        $g_crossrefs{$label} = "#$label";
+        $self->{_crossrefs}{$label} = "#$label";
         $whole_match;
     }xsge;
 
